@@ -1,5 +1,6 @@
 #![allow(dead_code)]
 #![warn(clippy::perf)]
+#![feature(unchecked_math)]
 #![warn(clippy::complexity)]
 #![feature(core_intrinsics)]
 #![warn(clippy::correctness)]
@@ -17,7 +18,7 @@ type fb = framebuffer;
 pub struct framebuffer {
   pub width: usize,
   pub height: usize,
-  ptr: (bool, *mut ()),
+  ptr: (bool, *mut u8),
 }
 
 impl Drop for framebuffer {
@@ -33,16 +34,20 @@ impl Clone for framebuffer {
 }
 
 impl framebuffer {
-  pub fn new(width: usize, height: usize) -> Self { return Self { width, height, ptr: (true, alloc(4 * width * height) as _) }; }
-  pub const unsafe fn from_ptr(width: usize, height: usize, ptr: *mut u8) -> Self { return Self { width, height, ptr: (false, ptr as _) }; }
+  pub fn new(width: usize, height: usize) -> Self { return Self { width, height, ptr: (true, alloc(4 * width * height)) }; }
+  pub const unsafe fn from_ptr(width: usize, height: usize, ptr: *mut u8) -> Self { return Self { width, height, ptr: (false, ptr) }; }
   pub const unsafe fn from(width: usize, height: usize, buffer: &[u8]) -> Self { return Self { width, height, ptr: (false, buffer.as_ptr() as _) }; }
-  pub unsafe fn from_mut(width: usize, height: usize, buffer: &mut [u8]) -> Self { return Self { width, height, ptr: (false, buffer.as_mut_ptr() as _) }; }
+  pub unsafe fn from_mut(width: usize, height: usize, buffer: &mut [u8]) -> Self { return Self { width, height, ptr: (false, buffer.as_mut_ptr()) }; }
 
   pub fn ptr_mut<T>(&mut self) -> *mut T { return self.ptr.1 as *mut T; }
   pub const fn ptr<T>(&self) -> *const T { return self.ptr.1 as *const T; }
   pub const fn len(&self) -> usize { return 4 * self.width * self.height; }
   pub fn slice<T>(&self) -> &[T] { return unsafe { std::slice::from_raw_parts(self.ptr(), self.len() / std::mem::size_of::<T>()) }; }
+  pub fn get(&self, x: usize, y: usize) -> &[u8] { return unsafe { std::slice::from_raw_parts(self.ptr::<u8>().add(4 * (x + y * self.width)), 4) }; }
   pub fn slice_mut<T>(&mut self) -> &mut [T] { return unsafe { std::slice::from_raw_parts_mut(self.ptr_mut(), self.len() / std::mem::size_of::<T>()) }; }
+  pub fn get_mut(&mut self, x: usize, y: usize) -> &mut [u8] { return unsafe { std::slice::from_raw_parts_mut(self.ptr_mut::<u8>().add(4 * (x + y * self.width)), 4) }; }
+  pub fn set(&mut self, x: usize, y: usize, r: u8, g: u8, b: u8, a: u8) { unsafe { *self.ptr_mut::<u32>().add(x + y * self.width) = r as u32 | ((g as u32) << 8) | ((b as u32) << 16) | ((a as u32) << 24); }; }
+  pub fn vec<T>(mut self) -> Vec<T> { self.ptr.0 = false; return unsafe { Vec::from_raw_parts(self.ptr.1 as *mut T, self.len() / std::mem::size_of::<T>(), std::alloc::Layout::from_size_align_unchecked(self.len(), 16).size() / std::mem::size_of::<T>()) }; }
 }
 
 pub mod ops {
@@ -98,6 +103,45 @@ pub mod ops {
     }
   }
 
+  pub mod rotate {
+    use super::*;
+
+    pub fn rotate180(fb: &mut fb) {
+      fb.slice_mut::<u32>().reverse();
+    }
+
+    pub fn rotate90(fb: &mut fb) {
+      let width = fb.width;
+      let height = fb.height;
+      let u32 = fb.ptr_mut::<u32>();
+      let o32 = fb.clone().ptr::<u32>();
+
+      fb.width = height;
+      fb.height = width;
+
+      for y in 0..height {
+        let yoffset = y * width;
+        let heighty1 = height - 1 - y;
+        for x in 0..width { unsafe { *u32.add(heighty1 + x * height) = *o32.add(x + yoffset); } }
+      }
+    }
+
+    pub fn rotate270(fb: &mut fb) {
+      let width = fb.width;
+      let height = fb.height;
+      let u32 = fb.ptr_mut::<u32>();
+      let o32 = fb.clone().ptr::<u32>();
+
+      fb.width = height;
+      fb.height = width;
+
+      for y in 0..height {
+        let yoffset = y * width;
+        for x in 0..width { unsafe { *u32.add(y + height * (width - 1 - x)) = *o32.add(x + yoffset); } }
+      }
+    }
+  }
+
   pub mod overlay {
     use super::*;
 
@@ -187,7 +231,7 @@ pub mod ops {
     pub fn nearest(fb: &fb, width: usize, height: usize) -> fb {
       let owidth = fb.width;
       let oheight = fb.height;
-      let old = fb.ptr::<u32>();
+      let o32 = fb.ptr::<u32>();
       let mut fb = framebuffer::new(width, height);
 
       let u32 = fb.ptr_mut::<u32>();
@@ -197,7 +241,7 @@ pub mod ops {
       for y in 0..height {
         let yoffset = y * width;
         let yyoffset = owidth * (yw * y as f64) as usize;
-        for x in 0..width { unsafe { *u32.add(x + yoffset) = *old.add(yyoffset + (xw * x as f64) as usize); } };
+        for x in 0..width { unsafe { *u32.add(x + yoffset) = *o32.add(yyoffset + (xw * x as f64) as usize); } };
       };
 
       return fb;
@@ -206,7 +250,7 @@ pub mod ops {
     pub fn linear(fb: &fb, width: usize, height: usize) -> fb {
       let owidth = fb.width;
       let oheight = fb.height;
-      let old = fb.ptr::<u8>();
+      let o8 = fb.ptr::<u8>();
       let mut fb = framebuffer::new(width, height);
 
       let mut offset = 0;
@@ -232,10 +276,10 @@ pub mod ops {
           let s3 = clamped(1 + xxi, 1 + yyi, owidth, oheight);
 
           unsafe {
-            *u8.offset(offset) = lerp(lerp(*old.add(s0), *old.add(s2), tx), lerp(*old.add(s1), *old.add(s3), tx), ty);
-            *u8.offset(1 + offset) = lerp(lerp(*old.add(1 + s0), *old.add(1 + s2), tx), lerp(*old.add(1 + s1), *old.add(1 + s3), tx), ty);
-            *u8.offset(2 + offset) = lerp(lerp(*old.add(2 + s0), *old.add(2 + s2), tx), lerp(*old.add(2 + s1), *old.add(2 + s3), tx), ty);
-            *u8.offset(3 + offset) = lerp(lerp(*old.add(3 + s0), *old.add(3 + s2), tx), lerp(*old.add(3 + s1), *old.add(3 + s3), tx), ty);
+            *u8.offset(offset) = lerp(lerp(*o8.add(s0), *o8.add(s2), tx), lerp(*o8.add(s1), *o8.add(s3), tx), ty);
+            *u8.offset(1 + offset) = lerp(lerp(*o8.add(1 + s0), *o8.add(1 + s2), tx), lerp(*o8.add(1 + s1), *o8.add(1 + s3), tx), ty);
+            *u8.offset(2 + offset) = lerp(lerp(*o8.add(2 + s0), *o8.add(2 + s2), tx), lerp(*o8.add(2 + s1), *o8.add(2 + s3), tx), ty);
+            *u8.offset(3 + offset) = lerp(lerp(*o8.add(3 + s0), *o8.add(3 + s2), tx), lerp(*o8.add(3 + s1), *o8.add(3 + s3), tx), ty);
           }
 
           offset += 4;
@@ -248,7 +292,7 @@ pub mod ops {
     pub fn cubic(fb: &fb, width: usize, height: usize) -> fb {
       let owidth = fb.width;
       let oheight = fb.height;
-      let old = fb.ptr::<u8>();
+      let o8 = fb.ptr::<u8>();
       let mut fb = framebuffer::new(width, height);
 
       let mut offset = 0;
@@ -293,25 +337,25 @@ pub mod ops {
           let s15 = clamped(2 + xxi, 2 + yyi, owidth, oheight);
 
           unsafe {
-            let c0 = hermite(*old.add(s0) as f64, *old.add(s1) as f64, *old.add(s2) as f64, *old.add(s3) as f64, tx);
-            let c00 = hermite(*old.add(1 + s0) as f64, *old.add(1 + s1) as f64, *old.add(1 + s2) as f64, *old.add(1 + s3) as f64, tx);
-            let c000 = hermite(*old.add(2 + s0) as f64, *old.add(2 + s1) as f64, *old.add(2 + s2) as f64, *old.add(2 + s3) as f64, tx);
-            let c0000 = hermite(*old.add(3 + s0) as f64, *old.add(3 + s1) as f64, *old.add(3 + s2) as f64, *old.add(3 + s3) as f64, tx);
+            let c0 = hermite(*o8.add(s0) as f64, *o8.add(s1) as f64, *o8.add(s2) as f64, *o8.add(s3) as f64, tx);
+            let c00 = hermite(*o8.add(1 + s0) as f64, *o8.add(1 + s1) as f64, *o8.add(1 + s2) as f64, *o8.add(1 + s3) as f64, tx);
+            let c000 = hermite(*o8.add(2 + s0) as f64, *o8.add(2 + s1) as f64, *o8.add(2 + s2) as f64, *o8.add(2 + s3) as f64, tx);
+            let c0000 = hermite(*o8.add(3 + s0) as f64, *o8.add(3 + s1) as f64, *o8.add(3 + s2) as f64, *o8.add(3 + s3) as f64, tx);
 
-            let c1 = hermite(*old.add(s4) as f64, *old.add(s5) as f64, *old.add(s6) as f64, *old.add(s7) as f64, tx);
-            let c11 = hermite(*old.add(1 + s4) as f64, *old.add(1 + s5) as f64, *old.add(1 + s6) as f64, *old.add(1 + s7) as f64, tx);
-            let c111 = hermite(*old.add(2 + s4) as f64, *old.add(2 + s5) as f64, *old.add(2 + s6) as f64, *old.add(2 + s7) as f64, tx);
-            let c1111 = hermite(*old.add(3 + s4) as f64, *old.add(3 + s5) as f64, *old.add(3 + s6) as f64, *old.add(3 + s7) as f64, tx);
+            let c1 = hermite(*o8.add(s4) as f64, *o8.add(s5) as f64, *o8.add(s6) as f64, *o8.add(s7) as f64, tx);
+            let c11 = hermite(*o8.add(1 + s4) as f64, *o8.add(1 + s5) as f64, *o8.add(1 + s6) as f64, *o8.add(1 + s7) as f64, tx);
+            let c111 = hermite(*o8.add(2 + s4) as f64, *o8.add(2 + s5) as f64, *o8.add(2 + s6) as f64, *o8.add(2 + s7) as f64, tx);
+            let c1111 = hermite(*o8.add(3 + s4) as f64, *o8.add(3 + s5) as f64, *o8.add(3 + s6) as f64, *o8.add(3 + s7) as f64, tx);
 
-            let c2 = hermite(*old.add(s8) as f64, *old.add(s9) as f64, *old.add(s10) as f64, *old.add(s11) as f64, tx);
-            let c22 = hermite(*old.add(1 + s8) as f64, *old.add(1 + s9) as f64, *old.add(1 + s10) as f64, *old.add(1 + s11) as f64, tx);
-            let c222 = hermite(*old.add(2 + s8) as f64, *old.add(2 + s9) as f64, *old.add(2 + s10) as f64, *old.add(2 + s11) as f64, tx);
-            let c2222 = hermite(*old.add(3 + s8) as f64, *old.add(3 + s9) as f64, *old.add(3 + s10) as f64, *old.add(3 + s11) as f64, tx);
+            let c2 = hermite(*o8.add(s8) as f64, *o8.add(s9) as f64, *o8.add(s10) as f64, *o8.add(s11) as f64, tx);
+            let c22 = hermite(*o8.add(1 + s8) as f64, *o8.add(1 + s9) as f64, *o8.add(1 + s10) as f64, *o8.add(1 + s11) as f64, tx);
+            let c222 = hermite(*o8.add(2 + s8) as f64, *o8.add(2 + s9) as f64, *o8.add(2 + s10) as f64, *o8.add(2 + s11) as f64, tx);
+            let c2222 = hermite(*o8.add(3 + s8) as f64, *o8.add(3 + s9) as f64, *o8.add(3 + s10) as f64, *o8.add(3 + s11) as f64, tx);
 
-            let c3 = hermite(*old.add(s12) as f64, *old.add(s13) as f64, *old.add(s14) as f64, *old.add(s15) as f64, tx);
-            let c33 = hermite(*old.add(1 + s12) as f64, *old.add(1 + s13) as f64, *old.add(1 + s14) as f64, *old.add(1 + s15) as f64, tx);
-            let c333 = hermite(*old.add(2 + s12) as f64, *old.add(2 + s13) as f64, *old.add(2 + s14) as f64, *old.add(2 + s15) as f64, tx);
-            let c3333 = hermite(*old.add(3 + s12) as f64, *old.add(3 + s13) as f64, *old.add(3 + s14) as f64, *old.add(3 + s15) as f64, tx);
+            let c3 = hermite(*o8.add(s12) as f64, *o8.add(s13) as f64, *o8.add(s14) as f64, *o8.add(s15) as f64, tx);
+            let c33 = hermite(*o8.add(1 + s12) as f64, *o8.add(1 + s13) as f64, *o8.add(1 + s14) as f64, *o8.add(1 + s15) as f64, tx);
+            let c333 = hermite(*o8.add(2 + s12) as f64, *o8.add(2 + s13) as f64, *o8.add(2 + s14) as f64, *o8.add(2 + s15) as f64, tx);
+            let c3333 = hermite(*o8.add(3 + s12) as f64, *o8.add(3 + s13) as f64, *o8.add(3 + s14) as f64, *o8.add(3 + s15) as f64, tx);
 
             *u8.offset(offset) = hermite(c0, c1, c2, c3, ty) as u8;
             *u8.offset(1 + offset) = hermite(c00, c11, c22, c33, ty) as u8;
@@ -347,163 +391,18 @@ pub mod ops {
 
       let width = fb.width;
       let height = fb.height;
-      unsafe { gcb(fb, width, height, cof); }
-    }
 
-    #[cfg(not(target_feature = "simd128"))]
-    unsafe fn gcb(fb: &mut fb, width: usize, height: usize, cof: (f32, f32, f32, f32, f32, f32, f32, f32)) {
-      let u8 = fb.ptr_mut::<u8>();
-      let o8 = alloc(fb.len());
-      let f32 = alloc(4 * 4 * width.max(height)) as *mut f32;
+      unsafe {
+        let o8 = alloc(fb.len());
+        let u8 = fb.ptr_mut::<u8>();
+        let f32 = alloc(4 * 4 * width.max(height)) as *mut f32;
 
-      u8.copy_to(o8, fb.len());
-      gc(o8, u8, f32, width, height, cof);
-      gc(u8, o8, f32, height, width, cof);
+        u8.copy_to(o8, fb.len());
+        gc(o8, u8, f32, width, height, cof);
+        gc(u8, o8, f32, height, width, cof);
 
-      free(o8, fb.len());
-      free(f32 as *mut u8, 4 * 4 * width.max(height));
-    }
-
-    #[cfg(target_arch = "wasm32")]
-    #[cfg(target_feature = "simd128")]
-    unsafe fn gcb(fb: &mut fb, width: usize, height: usize, cof: (f32, f32, f32, f32, f32, f32, f32, f32)) {
-      let u8 = fb.ptr_mut::<u8>();
-      let n32 = ffi::mem::alloc(4 * fb.len()) as *mut f32;
-
-      for y in 0..height {
-        let yoffset = y * width;
-
-        for x in 0..width {
-          let offset = 4 * (x + yoffset);
-          *n32.add(offset) = *u8.add(offset) as f32;
-          *n32.add(1 + offset) = *u8.add(1 + offset) as f32;
-          *n32.add(2 + offset) = *u8.add(2 + offset) as f32;
-          *n32.add(3 + offset) = *u8.add(3 + offset) as f32;
-        }
-      }
-
-      let o32 = ffi::mem::alloc(4 * fb.len()) as *mut f32;
-      let f32 = ffi::mem::alloc(4 * 4 * width.max(height)) as *mut f32;
-
-      n32.copy_to(o32, fb.len());
-      gcs(o32, n32, f32, width, height, cof);
-      gcse(u8, o32, f32, height, width, cof);
-      ffi::mem::free(n32 as *mut u8, 4 * fb.len());
-      ffi::mem::free(o32 as *mut u8, 4 * fb.len());
-      ffi::mem::free(f32 as *mut u8, 4 * 4 * width.max(height));
-    }
-
-    #[cfg(target_arch = "wasm32")]
-    #[cfg(target_feature = "simd128")]
-    unsafe fn gcs(n32: *mut f32, o32: *mut f32, f32: *mut f32, width: usize, height: usize, (k, a1, a2, a3, b1, b2, lc, rc): (f32, f32, f32, f32, f32, f32, f32, f32)) {
-      use std::arch::wasm32::*;
-
-      let k = f32x4_splat(k);
-      let a1 = f32x4_splat(a1);
-      let a2 = f32x4_splat(a2);
-      let a3 = f32x4_splat(a3);
-      let b1 = f32x4_splat(b1);
-      let b2 = f32x4_splat(b2);
-      let lc = f32x4_splat(lc);
-      let rc = f32x4_splat(rc);
-
-      let width4 = 4 * width;
-      let height4 = 4 * height;
-      let hw1 = height * (width - 1);
-
-      for y in 0..height {
-        let mut toffset = 0;
-        let mut ooffset = y * width4;
-        let mut offset = 4 * (y + hw1);
-        let mut po = v128_load(o32.add(ooffset) as *mut v128); let mut fu = f32x4_mul(lc, po); let mut tu = fu;
-
-        for _ in 0..width {
-          let co = v128_load(o32.add(ooffset) as *mut v128);
-          let cu = f32x4_add(f32x4_add(f32x4_mul(k, co), f32x4_mul(a1, po)), f32x4_add(f32x4_mul(b1, fu), f32x4_mul(b2, tu)));
-
-          tu = fu; fu = cu; po = co;
-          v128_store(f32.offset(toffset) as *mut v128, fu);
-
-          ooffset += 4;
-          toffset += 4;
-        }
-
-        ooffset -= 4;
-        toffset -= 4;
-
-        let mut po = v128_load(o32.add(ooffset) as *mut v128); tu = f32x4_mul(rc, po); fu = tu; let mut co = po;
-
-        for _ in 0..width {
-          let cu = f32x4_add(f32x4_add(f32x4_mul(a2, co), f32x4_mul(a3, po)), f32x4_add(f32x4_mul(b1, fu), f32x4_mul(b2, tu)));
-
-          tu = fu; fu = cu; po = co;
-          co = v128_load(o32.add(ooffset) as *mut v128);
-          v128_store(n32.add(offset) as _, f32x4_add(fu, v128_load(f32.offset(toffset) as *mut v128)));
-
-          ooffset -= 4;
-          toffset -= 4;
-          offset -= height4;
-        }
-      }
-    }
-
-    #[cfg(target_arch = "wasm32")]
-    #[cfg(target_feature = "simd128")]
-    unsafe fn gcse(u8: *mut u8, o32: *mut f32, f32: *mut f32, width: usize, height: usize, (k, a1, a2, a3, b1, b2, lc, rc): (f32, f32, f32, f32, f32, f32, f32, f32)) {
-      use std::arch::wasm32::*;
-      use std::intrinsics::{float_to_int_unchecked as fi};
-
-      let k = f32x4_splat(k);
-      let a1 = f32x4_splat(a1);
-      let a2 = f32x4_splat(a2);
-      let a3 = f32x4_splat(a3);
-      let b1 = f32x4_splat(b1);
-      let b2 = f32x4_splat(b2);
-      let lc = f32x4_splat(lc);
-      let rc = f32x4_splat(rc);
-
-      let width4 = 4 * width;
-      let height4 = 4 * height;
-      let hw1 = height * (width - 1);
-
-      for y in 0..height {
-        let mut toffset = 0;
-        let mut ooffset = y * width4;
-        let mut offset = 4 * (y + hw1);
-        let mut po = v128_load(o32.add(ooffset) as *mut v128); let mut fu = f32x4_mul(lc, po); let mut tu = fu;
-
-        for _ in 0..width {
-          let co = v128_load(o32.add(ooffset) as *mut v128);
-          let cu = f32x4_add(f32x4_add(f32x4_mul(k, co), f32x4_mul(a1, po)), f32x4_add(f32x4_mul(b1, fu), f32x4_mul(b2, tu)));
-
-          tu = fu; fu = cu; po = co;
-          v128_store(f32.offset(toffset) as *mut v128, fu);
-
-          ooffset += 4;
-          toffset += 4;
-        }
-
-        ooffset -= 4;
-        toffset -= 4;
-
-        let mut po = v128_load(o32.add(ooffset) as *mut v128); tu = f32x4_mul(rc, po); fu = tu; let mut co = po;
-
-        for _ in 0..width {
-          let cu = f32x4_add(f32x4_add(f32x4_mul(a2, co), f32x4_mul(a3, po)), f32x4_add(f32x4_mul(b1, fu), f32x4_mul(b2, tu)));
-
-          tu = fu; fu = cu; po = co;
-          co = v128_load(o32.add(ooffset) as *mut v128);
-          let f = &mut f32x4_add(fu, v128_load(f32.offset(toffset) as *mut v128)) as *mut v128 as *mut f32;
-
-          *u8.add(offset) = fi(*f);
-          *u8.add(1 + offset) = fi(*f.offset(1));
-          *u8.add(2 + offset) = fi(*f.offset(2));
-          *u8.add(3 + offset) = fi(*f.offset(3));
-
-          ooffset -= 4;
-          toffset -= 4;
-          offset -= height4;
-        }
+        free(o8, fb.len());
+        free(f32 as *mut u8, 4 * 4 * width.max(height));
       }
     }
 
